///////////////////////////////////////////////////////////////////////////////
//	モジュール : Smap.js, 2015-04-26, 西岡 芳晴 ( NISHIOKA Yoshiharu ),
//		Smap名前空間，定数・汎用関数用
//		S-map内では最初にロードされる必要がある．
///////////////////////////////////////////////////////////////////////////////

'use strict';

// Smap : 名前空間，S-map全体を表す名前空間
// GSJ, GSJMap : 下位互換のための別名．ただしすでに定義されているときは使用しない
// 	※GSJ, GSJMapは廃止予定
var
	Smap = { engine: {} },
	GSJMap,
	GSJ;

if ( typeof GSJ === 'undefined' ) {
	GSJ = Smap;
}
if ( typeof GSJMap === 'undefined' ) {
	GSJMap = Smap;
}

( function( Smap ) {

//
//	定数: 
//
	Smap.DEFAULT_BASE = {			// デフォルトのベースマップ，地理院地図(標準)
		id: 'CHIRIIN_STD',
		url: 'http://cyberjapandata.gsi.go.jp/xyz/{type}/{z}/{x}/{y}.png',
		title: '地理院地図(標準)',
		attribution: '国土地理院',
		attributionUrl: 'http://portal.cyberjapan.jp/help/termsofuse.html',
		attribution_url: 'http://portal.cyberjapan.jp/help/termsofuse.html', // 廃止予定
		minZoom: 0,
		maxZoom: 18,
		type: 'std',
	};
	Smap.DEFAULT_ELEVATION
			= 'http://gsj-seamless.jp/labs/elev/{type}/{z}/{y}/{x}.png?errorpng=on';
											// デフォルト標高レイヤーのURL
	Smap.DEFAULT_ENGINE 		= 'leaflet';	// デフォルトエンジン名
	Smap.DEFAULT_GEOCODER 	= 'csis';		// デフォルトジオコーダー
	Smap.EARTH_R = 6378137;			// 地球の長半径（単位ｍ），WGS84
	Smap.ZOOM_LIMIT = 24;			// このライブラリで利用できるズームレベルの限界値
	Smap.URL_RE =					// URL確認用正規表現
		/^(https?|ftp):\/\/[-_.!~*\'()a-zA-Z0-9\/?:\@&=+\$,%#{}]+$/; 
	Smap.BLANK_MAP = {
		id: 'BLANK_MAP',
		title: '無地の背景',
    	minZoom: 0,
    	maxZoom: Smap.ZOOM_LIMIT,
    	fillStyle: '#ffffff',
		drawTile: function( canvas, coord ) {
	        var
	        	ctx =canvas.getContext( '2d' );
	        ctx.fillStyle = this.fillStyle;
	        ctx.fillRect( 0, 0, 256, 256 );
		}
	};

//
//	関数: Smap.latLng()
//		LatLngオブジェクトの生成
//
	Smap.latLng = function( aPoint ) {
		var
			newPoint = null;
		if ( arguments.length === 1 ) {
			if ( typeof aPoint === 'string' ) {
				aPoint = aPoint.split( ',' );
			};
			if ( aPoint instanceof Array && aPoint.length > 1 ){
				aPoint = { lat: aPoint[ 0 ], lng: aPoint[ 1 ] };
			}
		} else if ( arguments.length > 1 ) {
			aPoint = { lat: arguments[ 0 ], lng: arguments[ 1 ] };
		}

		if ( aPoint && ( typeof aPoint.lat !== 'undefined' ) && ( typeof aPoint.lng !== 'undefined' ) ) {
			newPoint = { lat: Number( aPoint.lat ), lng: Number( aPoint.lng ) };
		}
		return newPoint
	}

//
//	関数: Smap.point() 
//		Pointオブジェクトの生成
//
	Smap.point = function( aPoint ) {
		var
			newPoint = null;

		if ( arguments.length === 1 ) {
			if ( typeof aPoint === 'string' ) {
				aPoint = aPoint.split( ',' );
			};
			if ( aPoint instanceof Array && aPoint.length > 1 ){
				aPoint = { x: aPoint[ 0 ], y: aPoint[ 1 ] };
			}
		} else if ( arguments.length > 1 ) {
			aPoint = { x: arguments[ 0 ], y: arguments[ 1 ] };
		}
		newPoint ={ x: Number( aPoint.x ), y: Number( aPoint.y ) };
		return newPoint;
	}

//
//	関数: Smap.bounds() 
//		Boundsオブジェクトの生成
//
	Smap.bounds = function( bounds ) {
		var
			newBounds = null;

		if ( arguments.length > 1 ) {
			bounds =  Array.prototype.slice.call( arguments );
		}
		if ( bounds instanceof Array ) {
			if ( bounds.length > 1 && bounds.length < 4 ){
				bounds = { sw: bounds[ 0 ], ne: bounds[ 1 ] };
			} else{
				bounds = { sw: [ bounds[ 0 ], bounds[ 1 ] ], ne: [ bounds[ 2 ], bounds[ 3 ] ] };
			};
		}
		if ( bounds && bounds.sw && bounds.ne ) {
			newBounds = { sw: Smap.latLng( bounds.sw ), ne: Smap.latLng( bounds.ne ) }
		};
		return newBounds;
	}

//
//	関数: Smap.latLngToPoint()
//		緯度経度をポイント座標に変換します
//		ここでのポイント座標とは，そのズームレベルにおいて256ピクセルのタイルを使用した場合の座標です．
//		左上を原点，右及び下をX, Y座標の正としたものです
//
	Smap.latLngToPoint = function( aLatLng, aZoom ) {
		var
			ll = Smap.latLng( aLatLng ),
			w = Math.pow( 2, aZoom ) * 128,
			yrad = Math.log( Math.tan( Math.PI * ( 90 + ll.lat ) / 360) );
		return { x: Math.round( ( ll.lng / 180 + 1 ) * w ), y: Math.round( ( 1 - yrad / Math.PI) * w) };
	};

//
//	関数: Smap.pointToLatLng()
//		ポイント座標を緯度経度に変換します
//		ここでのポイント座標とは，そのズームレベルにおいて256ピクセルのタイルを使用した場合の座標です．
//		左上を原点，右及び下をX, Y座標の正としたものです
//
	Smap.pointToLatLng = function( aPoint, aZoom ) {
		var
			w = Math.pow( 2, aZoom ) * 256,	// 世界全体のピクセル幅
			yy = Math.atan( Math.pow( Math.E, ( 0.5 - aPoint.y / w ) * 2 * Math.PI ) );
		return { lat: yy / Math.PI * 360 - 90, lng: aPoint.x / w * 360 - 180 };
	};


//
//	関数: Smap.contains()
//		緯度経度範囲が位置を含むか否かを返します
//		矩形枠上の位置は含まれると判断します．
//		bounds.sw.lat <= bounds.ne.latでなければなりません
//		bounds.sw.lng > bounds.ne.lngの場合は子午線をはさんだ範囲となります
	Smap.contains = function( bounds, latLng ) {
		var
			lTest,
			b = Smap.bounds( bounds ),
			ll = Smap.latLng( latLng );

		if ( b.sw.lng <= b.ne.lng ) {
			lTest = b.sw.lng <= ll.lng && b.ne.lng >= ll.lng;
		} else {
			lTest = b.sw.lng >= ll.lng && b.ne.lng <= ll.lng;
		}
		return ( lTest && b.sw.lat <= ll.lat && b.ne.lat >= ll.lat );
	};

//
//	関数: Smap.zoomToScale()
//		ズームレベルをスケール(m/ピクセル)に変換します．
//
	Smap.zoomToScale = function( aZoom, aLat ){
		return 2 * Math.PI * Smap.EARTH_R * Math.cos( aLat / 180 * Math.PI ) / 256 / Math.pow( 2, aZoom )
	}

//
//	関数: Smap.scaleToZoom()
//		ズームレベルをスケール(m/ピクセル)に変換します．
//
	Smap.scaleToZoom = function( aScale, aLat ){
		var
			scale0 = 2// Math.PI * Smap.EARTH_R * Math.cos( aLat / 180 * Math.PI ) / 256;
		return Math.LOG2E * Math.log( scale0 / aScale )
	}

//
//	関数: Smap.zyToScale()
//		ズームレベルとピクセルY座標をスケール(m/ピクセル)に変換します．
//
	Smap.zoomYToScale = function( aZoom, aY ){
		var
			w = Math.pow( 2, aZoom ) * 256,	// 世界全体のピクセル幅
			yy = Math.atan( Math.pow( Math.E, ( 0.5 - aY / w ) * 2 * Math.PI ) );
		return 2 * Math.PI * Smap.EARTH_R * Math.sin( yy * 2 ) / w
	}

//
//	関数: Smap.elevationLayer()
//		ElevationLayerオブジェクトを生成します．
//
	Smap.elevationLayer = function( url ){
		var
			elevationLayer = {};

		elevationLayer.url = ( url ) ? url : Smap.DEFAULT_ELEVATION;
		elevationLayer.getElevation = function( aLatLng, aZ, params ) {
			var
				point = Smap.latLngToPoint( Smap.latLng( aLatLng ), aZ ),
				x = Math.floor( point.x / 256 ),
				y = Math.floor( point.y / 256 ),
				i = point.x % 256,
				j = point.y % 256,
				url,
				type = 'm',
				invalid = Math.pow( 2, 23 ),
				promise,
				p2;

			if ( params && ( params.type === 'cm' ) ){
				type = 'cm';
			}
			
			url = this.url.replace( '{type}', type ).replace( '{z}', aZ )
						.replace( '{y}', y ).replace( '{x}', x );

			p2 = Smap.promise( function( resolve, reject ) {
				if ( params && params.pngzip ) {
					Smap.loadPngZip( url.replace( '.png', '.zip' ) ).then( function( img ) {
						var
							canvas = document.createElement( 'canvas' ),
							ctx = canvas.getContext( '2d' ),
							data,
							x;
						if ( i < 0 || i > img.width || j < 0 || j > img.height ) {
							resolve( null );
						} else {
							canvas.width = img.width;
							canvas.height = img.height;
							canvas.globalCompositeOperation = 'copy';	// 不透明度を正確に取得するため
							ctx.drawImage( img, 0, 0 );
							data = ctx.getImageData( i, j, 1, 1 ).data;
							resolve( data );
						}
					}, function( error ){ error } );	// loadPngZip のエラー時
				} else {
					Smap.loadImagePixel( url, i, j ).then( function( data ){
						resolve( data );
					}, function( error ){ error } );	// loadImagePixelのエラー時
				}
			} );

			promise = Smap.promise( function( resolve, reject ) {
				p2.then( function( data ) {
					var
						x,
						h;
					if ( data == null ) {
						reject();
					} else {
						x = data[0] * 256 * 256 + data[1] * 256 + data[2];
						if ( x === invalid ) {
							reject();
						} else {
							h = ( x < invalid ) ? x : x - invalid * 2;
							resolve( ( type === 'm' ) ? h : h / 100 );
						}
					}
				}, function(){} );	//標高タイルロード失敗時
			} );
			return promise;
		}
		return elevationLayer;
	}

//
//	関数: Smap.getElevation()
//		標高所得プロミスを発行します
//
	Smap.getElevation = function( aLatLng, aZ, params ) {
		return Smap.elevationLayer().getElevation( aLatLng, aZ, params );
	}

//
//	関数: Smap.getEngineNames()
//		使用できるマップエンジン名のリストを取得します．
//
	Smap.getEngineNames = function() {
		var
			engines = [];

		for ( var engine in Smap.engine ) {
			engines.push( engine );
		}
		return engines;
	}

//
//	関数: Smap.promoise()
//		プリミティブなプロミスオブジェクトの生成
//
	Smap.promise = function( starter ) {
		if ( typeof starter === 'undefined' ) {
			return Smap.promise( function( resolve ) { resolve(); } );
		}
		if ( typeof Promise !== 'undefined' ) { // Promiseが実装されて入ればそれを使用する
			return new Promise( starter );
		}
		var p = {
			state: 0, // 0: 未完了，1: 完了，2: 拒否
			success: [],
			error: [],
			then: function( onSuccess, onError ) {
				var
					that = this,
					p2;

				p2 = Smap.promise( function( resolve, reject ) {
					that.success.push( function( value ) {
						var
							v;
						if ( onSuccess ) {
							// 2015-03-17 この付近のアルゴリズムがおかしい．
							// エラーが正しく伝搬しない
							if ( value !== null && typeof value !== 'undefined'
									&& value.then ){
								value.then( function( v ){
//									console.log( onSuccess( v ) );
									resolve( onSuccess( v ) );
								}, function( error ) {
									reject( error );
								} );
							} else {
								v = onSuccess( value );
								if ( v !== null && typeof v !=='undefined' && v.then ) {
									v.then( function( v2 ) {
										resolve( v2 );
									}, function( error ){
										reject( error );
									} );
								} else {
									resolve( v );
								}
//								resolve( v );
							}
						} else {
							resolve();
						}
					} );
					if ( onSuccess && ( that.state == 1 ) ){
						if ( that.resolveValue !== null &&
								typeof that.resolveValue !== 'undefined' &&
								typeof that.resolveValue.then !== 'undefined' ){
							that.resolveValue.then( function( v){
//								console.log( onSuccess( that.resolveValue ) );
								resolve( onSuccess( v ) );
							} );
						} else {
//							console.log( onSuccess( that.resolveValue ) );
							resolve( onSuccess( that.resolveValue ) );
						}
					}
					that.error.push( function( value ) {
						if ( onError ) {
							resolve( onError( value ) );
//							reject( onError( value ) );
						} else {
//							resolve();
							reject( value );
						}
					} );
					if ( onError && ( that.state == 2 ) ) {
//						resolve( onError( that.errorValue ) );
						reject( onError( that.errorValue ) );
					}
				} );
				return p2;
			},
			resolve: function( value ) {
//				var
//					args = arguments;
				this.resolveValue = value;
				this.success.forEach( function( aSuccess ) {
//					aSuccess.apply( null, args )
					aSuccess( value )
				} );
				this.state = 1;
			},
//			reject: function() {
			reject: function( value ) {
//				var
//					args = arguments;
				this.errorValue = value;
				this.error.forEach( function( aError ) {
//					aError.apply( null, args );
					aError( value );
				} );
				this.state = 2;
			},
			catch : function( onRejected ){
				this.then( undefined, onRejected );
			}
		};
		if ( starter ) {
			setTimeout( starter(
				function( value ) { p.resolve( value ) },
				function( value ) { p.reject( value ) }
			), 0 );
		}
		return p
	}

//	関数: Smap.promise.resovle()
//		ただちに成功するプロミスを生成する
//
	Smap.promise.resolve = function( value ){
		var
			p;

		if ( typeof Promise !== 'undefined' ) { // Promiseが実装されて入ればそれを使用する
			p = Promise.resolve( value );
		} else {
			p =	Smap.promise( function( resolve ) {
				if ( typeof value !== 'undefined' && value.then ){	// thenableならば
					value.then( function( v ){
						resolve( v );
					} );
				} else {
					resolve( value );
				}
			} );
		}
		return p;
	};

//	関数: Smap.promise.reject()
//		ただちに失敗するプロミスを生成する
//
	Smap.promise.reject = function( value ){
		var
			p;

		if ( typeof Promise !== 'undefined' ) { // Promiseが実装されて入ればそれを使用する
			p = Promise.reject( value );
		}else {
			p =	Smap.promise( function( undefined, reject ) {
				reject( value );
			} );
		}
		return p;
	};

//	関数: Smap.promise.race()
//		引数にプロミスの配列を指定し，いずれかが最初に成功または失敗したときにその結果を返します
//		race()の返すプロミスが固定しても，プロミスリスト内のプロミスの実行は影響を受けません
	Smap.promise.race = function( aPromiseList ){
		var
			promise,
			state = 0;	//プロミスの状態 0: Pending

		promise = Smap.promise( function( resolve, reject ) {
			aPromiseList.forEach( function( aPromise ) {
				aPromise.then( function( value ){
					if ( state === 0 ){				// Pendingならば
						state = 1;					// Fulfilledに変更
						resolve( value );
					}
				}, function( error ) {
					if ( state === 0 ){		// Pendingならば
						state = 2;					// エラーに変更
						reject( error );
					}
				} );
			} );
		} );
		return promise;
	};

//	関数: Smap.promise.all()
//	関数: Smap.promiseAll() ... 廃止予定
//		すべての子プロミスが成功したときに成功となるプロミスの生成
//		すべて成功したとき，結果は全てのプロミスから得られた値の配列として渡されます
//		いずれかが失敗したとき、直ちに失敗したプロミスの結果をもって失敗となり，他の結果は無視されます
//
	Smap.promise.all = function ( aPromiseList ) {
		var
			count = aPromiseList.length,
			promise;

		if ( typeof aPromiseList.length === 'undefined' ) { // 引数が配列でないならば
			count = arguments.length;
			aPromiseList = [];
			for ( var i = 0; i < count; i++ ) {
				aPromiseList[i] = arguments[i];
			}
		}

		promise = Smap.promise( function( resolve, reject ) {
			var
				results = [];

			if ( count === 0 ) {
				resolve( results );
				return;
			};
			aPromiseList.forEach( function( aPromise, index ) {
				aPromise.then( function( data ) {
					results[ index ] = data;
					count--;
					if ( count == 0 ) {
						resolve( results );
					}
				}, function( error ) {
					reject( error );
				} );
			} );
		});

		return promise;
	}
	Smap.promiseAll =Smap.promise.all;	// Smap.promiseAllは廃止予定
//
//	関数: Smap.promise.finishing()
//	関数: Smap.promiseFinishing()	// 廃止予定
//		すべての子プロミスの終了を監視するプロミスの生成
//		各プロミスの完了/拒否は問わない．
//		このプロミス自体が拒否状態になることはない．
//		この仕様は，一つでも拒否状態になれば拒否を返すES6のallとはことなります．
//
	Smap.promise.finishing = function ( aPromiseList ) {
		var
			count = aPromiseList.length,
			promise;

		if ( typeof aPromiseList.length === 'undefined' ) { // 引数が配列でないならば
			count = arguments.length;
			aPromiseList = [];
			for ( var i = 0; i < count; i++ ) {
				aPromiseList[i] = arguments[i];
			}
		}

		promise = Smap.promise( function( resolve, reject ) {
			var
				results = [];

			if ( count === 0 ) {
				resolve( results );
				return;
			}
			aPromiseList.forEach( function( aPromise, index ) {
				aPromise.then( function( data ) {	// 成功時
					results[ index ] = data;
				}, function(){ 						// 失敗時
						// 本来ならこの行はいらないはずだが，ChromeのPromiseでは
						// 失敗時の処理を何も記述しないと，次のthenメソッドで指定した処理が
						// 実行されない
				} ).then( function() { // 成功・拒否かかわらず実行
					count--;
//					console.log( 'count' + count );
					if ( count === 0 ) {
						resolve( results );
					}
				} );
			} );
		});

		return promise;
	}
	Smap.promiseFinishing = Smap.promise.finishing;	// Smap.promiseFinishingは廃止予定
	Smap.allPromise = function ( aPromiseList ) {
		var
			count = aPromiseList.length,
			promise;

		if ( typeof aPromiseList.length === 'undefined' ) { // 引数が配列でないならば
			count = arguments.length;
			aPromiseList = [];
			for ( var i = 0; i < count; i++ ) {
				aPromiseList[i] = arguments[i];
			}
		}

		promise = Smap.promise( function( resolve, reject ) {
			aPromiseList.forEach( function( aPromise ) {
				aPromise.then().then( function() { // 成功・拒否かかわらず実行
					count--;
					if ( count == 0 ) {
						resolve();
					}
				} );
			} );
		});

		return promise;
	}

//
//	関数: Smap.loadImage()
//		画像ファイルのロードプロミスの生成
//
	Smap.loadImage = function( url, params ) {
		var
			img = new Image(),
			promise = Smap.promise( function( resolve, reject ) {
				// data:スキーマでない場合にのみ'anonymous'を設定
				// Firefoxではdata:スキーマで設定してしまうとCORSエラーで読み込めない
				if ( url.substring( 0, 5 ) !== 'data:' ) {	
					img.crossOrigin = 'anonymous';  		
				}
				if ( params ) {
					for ( var key in params ) {
						img[ key ] = params[ key];
					}
				}
				img.onload = function() {
					resolve( this );
				};
				img.onerror = function() {
					reject( { msg: 'Error : image load error ' } );
				}
				img.src = url;
			} );
		return promise;
	}

//
//	関数: Smap.loadImagePixel()
//		画像ファイル指定ピクセルのデータロードプロミスの生成
//
	Smap.loadImagePixel = function( aUrl, i, j ) {
		var
			promise = Smap.promise( function( resolve, reject ) {
				Smap.loadImage( aUrl ).then(
					function( img ) {
						var
							canvas = document.createElement( 'canvas' ),
							ctx = canvas.getContext( '2d' ),
							d;

						if ( i < 0 || i > img.width || j < 0 || j > img.height ) {
								resolve( null );
						} else {
							canvas.width = img.width;
							canvas.height = img.height;
							canvas.globalCompositeOperation = 'copy';	// 不透明度を正確に取得するため
							ctx.drawImage( img, 0, 0 );
							d = ctx.getImageData( i, j, 1, 1 ).data;
							resolve( Array.prototype.slice.apply( d ) );
						}
					},
					function( error ) {
						reject( error );
					}
				) 
			} )
		return promise;
	}

//
//	関数: Smap.loadUrl()
//		XMLHttpRequestを追加った汎用的なファイルロードプロミスの生成
//
	Smap.loadUrl = function( url, type, headers ) {
		var
			promise = Smap.promise( function( resolve, reject ) {
				var
					req = new XMLHttpRequest();

				req.onreadystatechange = function (){
					if ( req.readyState === 4 ) {
						if( req.status == 0 ){
							reject( { msg: 'Error : XHR Connection Error', statuscode: 0 } );
						} else {
							if( ( 200 <= req.status && req.status < 300 ) || ( req.status == 304 ) ){
								resolve( req.response );
							}else{
								reject( { msg: 'Error: Request Error', statuscode: req.status  } );
							}
						}
					}
				};
//				req.open( 'POST' , url );
//				2015-02-11 POSTだとキャッシュが効かないためGETに変更
//				問題が生じるようなら引数を追加してGETとPOSTを切り替えられるようにする
				req.open( 'GET' , url );
				for ( var key in headers ) {
					req.setRequestHeader( key, headers[ key ] );
				}
				// Firefoxでは以下の1文をより前方で行うとエラーを出さず動作が停止する．
				req.responseType = ( type ) ? type : '';
				req.send( 'send' );
			} );
		return promise;
	}

//
//	関数: Smap.loadXml()
//		XMLファイルのロードプロミスの生成
//
	Smap.loadXml = function( url, headers ) {
		return Smap.loadUrl( url, 'document', headers );
	}

//
//	関数: Smap.loadBlob()
//		XMLファイルのロードプロミスの生成
//
	Smap.loadBlob = function( url, headers ) {
		return Smap.loadUrl( url, 'blob', headers );
	}

//
//	関数: Smap.loadArrayBuffer()
//		XMLファイルのロードプロミスの生成
//
	Smap.loadArrayBuffer = function( url, headers ) {
		return Smap.loadUrl( url, 'arraybuffer', headers );
	}

//
//	関数: Smap.loadText()
//		テキストファイルのロードプロミスの生成
//
	Smap.loadText = function( url, headers ) {
		return Smap.loadUrl( url, 'text', headers );
	}

//
//	関数: Smap.loadJSON()
//		JSONファイルのロードプロミスの生成
//
	Smap.loadJSON = function( url, headers ) {
		var
			promise = Smap.promise( function( resolve, reject) {
				Smap.loadText( url, headers ).then(
					function( text ) {
						resolve( JSON.parse( text ) );
					},
					function( error ) {
						reject( error );
					}
				);
			});
		return promise;
	}

//
//	関数: Smap.loadCSV()
//		CSVファイルのロードプロミスの生成
//
	Smap.loadCSV = function( url, separator, headers ) {
		var
			promise = Smap.promise( function( resolve, reject ) {
				separator = ( separator ) ? separator : ',' ;
				Smap.loadText( url, headers ).then(
					function( text ) {
						var
		    				result = [];
						text.split( '\n' ).forEach( function( line ) {
							result.push( line.split( separator ) );
						} );
						resolve( result );
					},
					function( error ) {
						reject( error );
					}
				);
			});
		return promise;
	};

//
//	関数: Smap.loadPngZip()
//		PNGZIP(zip圧縮したPNGファイル)のロードプロミスの生成
//
	Smap.loadPngZip = function( url, headers ) {
		var
			promise = Smap.promise( function( resolve, reject ) {
				Smap.loadUrl( url, 'arraybuffer', headers ).then(
					function( data ) {
						var
							blob,
							zip,
							img = new Image();

						if( data.byteLength == 0) {
							reject( 'error data length = 0' );
						} else {
							zip = new JSZip( data );
							for ( var key in zip.files ) {
								blob = new Blob( [ zip.file( key ).asArrayBuffer() ], { type: 'image/png' } );
								img.onload = function( event ){
 									resolve( this );
 									// ここでrevokeObjectURLをコールしても
 									// resolveないのよ処理よりも先に実行されてしうまう．
//	 									URL.revokeObjectURL( this.src );
								};
								img.onerror = function(){
									reject( 'create img url error' );
								}
								img.src = URL.createObjectURL( blob );
							}
//							}
						}
					},
					function( error ) {
						reject( error );
					}
				);
			} );
		return promise;
	};

// *
// *　関数: Smap.saveURL()
// *　　　URLをファイルへ保存
// *
	Smap.saveURL = function( aUrl, aFileName) {
		if( navigator.msSaveOrOpenBlob ) {	// Internet Explorerの場合
			// 以下はdata:スキーマとbase64指定が前提
			var
		    	tmp = aUrl.split( ',' ),
		    	data = atob( tmp[1] ),
				mime = tmp[0].split( ':' )[1].split( ';' )[0],
				arr = new Uint8Array( data.length );

			for ( var i = 0; i < data.length; i++ ) {
				arr[i] = data.charCodeAt( i );
			}
			navigator.msSaveOrOpenBlob( new Blob( [arr], { type: mime } ), aFileName );
		} else {									// Internet Explorer以外
			var
				a = document.createElement( 'a' ),
				evt = document.createEvent( 'MouseEvents' );
			a.href = aUrl;
			a.download= aFileName;
			evt.initMouseEvent( 'click', true, true, window, 0, 0, 0, 0, 0, 
					false, false, false, false, 0, null );
			a.dispatchEvent( evt );
		}
	};

	Smap.saveBlob = function( blob, aFileName ){
		if( navigator.msSaveOrOpenBlob ) {	// Internet Explorerの場合
			navigator.msSaveOrOpenBlob( blob, aFileName );
		} else {
			var
				url = ( window.URL || window.webkitURL ).createObjectURL( blob );
			Smap.saveURL( url, aFileName );
		}
	}

// *
// *　関数: Smap.mapToKMZ()
// *　　　Canvasに描かれた地図をKMZへ保存
// *　　　JSZipが必要
// *
	Smap.mapToKMZ =function( aCanvas, bounds, aFileName, params ) {
		var
			template = '<?xml version="1.0" encoding="UTF-8"?>' + '\n'
					+ '<kml xmlns="http://www.opengis.net/kml/2.2">'
					+ '<Document><GroundOverlay><name>{%name}</name><description>{%description}</description>'
					+ '<Icon><href>map.png</href></Icon><LatLonBox>'
					+ '<south>{%s}</south><west>{%w}</west><north>{%n}</north><east>{%e}</east>'
					+ '</LatLonBox></GroundOverlay></Document></kml>',
			name = ( params && params.name ) ? params.name : '',
			description = ( params && params.description ) ? params.description : '',
			zip = new JSZip(),
			base64 = aCanvas.toDataURL().split( 'base64,' )[1],
			head = 'data:application/vnd.google-earth.kmz;base64,';

		template = template.replace( '{%name}', name )
			.replace( '{%description}', description )
			.replace( '{%n}', bounds.ne.lat )
			.replace( '{%s}', bounds.sw.lat )
			.replace( '{%e}', bounds.ne.lng )
			.replace( '{%w}', bounds.sw.lng );
		zip.file( 'map.kml', template );
	    zip.file( 'map.png', base64, { base64: true } );
		Smap.saveURL( head  + zip.generate() , aFileName );
	};

//
//	関数: Smap.geocoder()
//		ジオコーダオブジェクトの生成
//
	Smap.geocoder =function( geocoder ) {
		var
			newGeocoder =  { geocoders: [] };

		newGeocoder.geocode = function( addr, params ) {		// デフォルトのgeocode()処理
			var
				ps = [],
				p;
			this.geocoders.forEach( function( gc ) {
				if ( params && params.bounds ) {
					params.bounds = Smap.bounds( params.bounds );
				}
				ps.push( gc.geocode( addr, params ) );
			} );
			p = Smap.promiseFinishing( ps ).then( function( data ){
				var
					results;

				results = data.reduce( function( pv, cv ){
					Array.prototype.push.apply( pv, cv.results );
					return pv;
				}, [] );
				return( { results: results } );
			} );
//			p = ps[ 0 ];
			return p;
		}

		if ( typeof geocoder === 'undefined' || geocoder === null ){ // 引数がないなら，デフォルト検索
			newGeocoder = Smap.geocoder[ Smap.DEFAULT_GEOCODER ]();
			newGeocoder.id = Smap.DEFAULT_GEOCODER;
		} else if ( typeof geocoder === 'string' ) {	 		// 引数が文字列なら，ID
			if ( Smap.geocoder[ geocoder ] ) {
				newGeocoder = Smap.geocoder[ geocoder ]();
			}
			newGeocoder.id = geocoder;
		} else if ( typeof geocoder === 'function' ) { 			// 引数が関数なら，geocode()メソッド
			newGeocoder.geocode = geocoder;
		} else if ( typeof geocoder.length !== 'undefined' ) {	// 引数が配列なら，geocodersプロパティ
			geocoder.forEach( function( gc ) {
				this.geocoders.push( Smap.geocoder( gc ) );
			}, newGeocoder );
		} else {												// 引数はオブジェクトとみなして
			newGeocoder =geocoder;
		}

		newGeocoder.add = function( geocoder ) {
			this.geocoders.push( Smap.geocoder( geocoder ) );
		}

		return newGeocoder;
	}

//
//	オブジェクト: Smap.base64url()
//		mBase64関連のメソッドを管理するオブジェクト
//
	Smap.base64url = {
		encodeBaseStr : 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_',
		encode : function ( array ) {
					var result = '';
					for ( var i = 0; i < Math.ceil( array.length / 6 ); i++ ) {
						var x = 0;
						for ( var j = 0; j < 6; j++ ) {
							x = x * 2;
							if ( ( i * 6 + j < array.length ) && ( array[ i * 6 + j ] ) ) {
								x += 1;
							}
						}
						result += this.encodeBaseStr.charAt( x );
					}
					return result;
				},
		decode : function ( str ) {
					var result = [];
					for ( var i = 0; i < str.length; i++ ) {
						for ( var z = 5; z >= 0; z-- ) {
							result.push( ( this.encodeBaseStr.indexOf( str[ i ] ) >> z ) & 1 );
						}
					}
					return result;
				}
		
	};

} ) ( Smap );

